/*
 * vi:ts=4:tw=78:shiftwidth=4:expandtab
 * vim600:fdm=marker
 *
 * iconvert.c  -  simple iconv() wrapper routines, see comments in iconvert.h
 * for usage.
 *
 * Copyright (C) 2004 by Zhang Le <ejoy@users.sourceforge.net>
 * Begin       : 29-Jul-2004
 * Last Change : 29-May-2005.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "iconvert.h"

#ifndef ICONV_CONST
/*    #warning ICONV_CONST not defined in config.h?. guessing from platform*/
    #if defined(__CYGWIN__) || defined(__FreeBSD__) || defined(__NetBSD__) || defined(WIN32) /* mingw defines WIN32 */
    #define ICONV_CONST const
    #endif

    #if defined(__linux__) 
    #define ICONV_CONST 
    #endif
#endif /* ICONV_CONST */

#if defined (WORDS_BIGENDIAN)
    #if (SIZEOF_UCHAR_T == 2)
        #define UCS_INTERNAL "UCS-2BE"
    #elif (SIZEOF_UCHAR_T == 4)
        #define UCS_INTERNAL "UCS-4BE"
    #else
        #warning unknown uchar_t size
    #endif
#else /* default is LITTLE-ENDIAN */
    #if (SIZEOF_UCHAR_T == 2)
        #define UCS_INTERNAL "UCS-2LE"
    #elif (SIZEOF_UCHAR_T == 4)
        #define UCS_INTERNAL "UCS-4LE"
    #else
        #warning unknown uchar_t size
    #endif
#endif

#define ICONV_BUFFER_SIZE 4096*4*sizeof(uchar_t)

iconv_info* iconv_create(const char* encoding) {
    iconv_info* info;

    info = (iconv_info*) malloc(sizeof(iconv_info));
    assert(info);

    info->m_iconv_from_unicode = iconv_open(encoding, UCS_INTERNAL);
    if (info->m_iconv_from_unicode == (iconv_t)(-1)) {
        perror("iconv_create(encoding) fail");
        free(info);
        return 0;
    }

    info->m_iconv_to_unicode = iconv_open(UCS_INTERNAL, encoding);
    if (info->m_iconv_to_unicode == (iconv_t)(-1)) {
        perror("iconv_create(encoding) fail");
        free(info);
        return 0;
    }

    info->m_encoding = (char*)malloc(strlen(encoding) + 1);
    assert(info->m_encoding);
    strcpy(info->m_encoding, encoding);

    info->m_buffer = (char*)malloc(ICONV_BUFFER_SIZE);
    assert(info->m_buffer);

    return info;
}

int iconv_destory(iconv_info* info) {
    assert(info);
    iconv_close(info->m_iconv_from_unicode);
    iconv_close(info->m_iconv_to_unicode);
    free(info->m_buffer);
    free(info->m_encoding);
    free(info);
    return 0;
}

int iconv_from_unicode(iconv_info* info, const uchar_t* from, size_t len_from,
        char* to, size_t len_to) {
    size_t insize = 0;
    size_t outsize = 0;
    size_t ret;
    ICONV_CONST char* psrc = (ICONV_CONST char*)from;
    char* pdest = info->m_buffer;

    assert(from);
    assert(to);

    iconv(info->m_iconv_from_unicode, NULL, &insize, NULL, &outsize);

    insize = len_from * sizeof(uchar_t);
    outsize = ICONV_BUFFER_SIZE;

    ret = iconv(info->m_iconv_from_unicode, &psrc, &insize, &pdest, &outsize);

    if (ret == (size_t)(-1))
        return -1;
    else {
        size_t len = pdest - info->m_buffer;
        if (len + 1 > len_to) {
            /* output buffer overflow */
            return -2;
        } else {
            memcpy(to, info->m_buffer, len);
            to[len] = '\0';
            return len;
        }
    }
}

int iconv_to_unicode(iconv_info* info, const char* from, size_t len_from,
        uchar_t* to, size_t len_to) {
	size_t insize = 0;
	size_t outsize = 0;
    size_t ret;
    ICONV_CONST char* psrc = (ICONV_CONST char*)from;
    char* pdest = info->m_buffer;

    iconv(info->m_iconv_to_unicode, NULL, &insize, NULL, &outsize);

	insize = len_from;
	outsize = ICONV_BUFFER_SIZE;

    ret = iconv(info->m_iconv_to_unicode, &psrc, &insize, &pdest, &outsize);

	if (ret == (size_t)(-1)) {
        return -1;
    } else {
        size_t len = (uchar_t*)pdest - (uchar_t*)info->m_buffer;
        if (len + 1> len_to) {
            /* output buffer overflow */
            return -2;
        } else {
            memcpy(to, info->m_buffer, len * sizeof(uchar_t));
            to[len] = L'\0';
            return len;
        }
/*        to.assign((uchar_t*)m_buffer,(uchar_t*)pdest);*/
/*        to.assign((wchar_t*)m_buffer,(BUFFER_SIZE - outsize)/(sizeof(wchar_t)*2));*/
    }
}

/* return converted unicode string allocated with malloc(), or 0 if failed */
uchar_t* str_to_unicode(const char* str, const char* encoding) {
    size_t len;
    int ulen;
    uchar_t* buf;
    uchar_t* ustr;
    iconv_info* info;

    len = strlen(str);
    buf = (uchar_t*)malloc(sizeof(uchar_t)*(len+1));
    assert(buf);

    info = iconv_create(encoding);
    assert(info);
    ulen = iconv_to_unicode(info, str, len, buf, len + 1);
    if (ulen >= 0) {
        ustr = (uchar_t*)malloc(sizeof(uchar_t) * (ulen + 1));
        assert(ustr);
        ustrcpy(ustr, buf);
    } else {
        ustr = 0;
    }

    free(buf);
    iconv_destory(info);
    return ustr;
}

